/* AMBGRO - program for convering AMBER topology to GROMACS topology */
/* Filip Ryjacek 2002 */
/* BDR Modified according to GMX mailing list suggestion by Tomas Kubar for
   longer atom names.  %4d changed to %6d for applicable fprintf(grotop */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <math.h>
/* #include <mcheck.h> */

#include "ambconv.h"
#include "write_grotop.h"


// prnt bonds
void write_top_bon_pert(s_atop* atop);
void write_top_bon_notpert(s_atop* atop);

// print pairs in perturbed dihedrals
void write_top_pairs_pert(s_atop* atop);
void write_top_pairs_notpert(s_atop* atop);

// print angles
void write_top_angles_pert(s_atop* atop);
void write_top_angles_notpert(s_atop* atop);

// print dihedrals
void write_top_dihedrals_pert(s_atop* atop);
void write_top_dihedrals_notpert(s_atop* atop);



/* write bonded information */
void write_top_bon(s_atop* atop)
  {

    // BONDS
    fprintf(grotop, "\n [ bonds ]\n;   ai  aj   funct      b0          kb\n");
    // first write non-perturbed bonds
     write_top_bon_notpert(atop);

    // then write perturbed bonds (if any)
    if (atop->ifpert > 0) {
      if (invert == 0) {
        fprintf(grotop, ";   ai  aj   funct    B_b0        B_kb           A_b0        A_kb\n");
      }
      else {      
        fprintf(grotop, ";   ai  aj   funct    A_b0        A_kb           B_b0        B_kb\n");
      }
      write_top_bon_pert(atop);
      }
    
    
    // PAIRS
    fprintf(grotop, "\n [ pairs ]\n");
    // print all dihedrals except 'end group' ineractions

    write_top_pairs_notpert(atop);

    // then write perturbed pairs
    if (atop->ifpert > 0) {
      fprintf(grotop, "; pairs in perturbed part of the system\n");
      write_top_pairs_pert(atop);
      }

    /* ANGLES */
    fprintf(grotop, "\n[ angles ]\n; ai  aj  ak  funct      th0         cth\n");

    write_top_angles_notpert(atop);

    // then write perturbed angles
    if (atop->ifpert > 0) {
      if (invert == 0) {
        fprintf(grotop, "; ai  aj  ak  funct    B_th0       B_cth       A_th0       A_cth\n");
      } 
      else {
        fprintf(grotop, "; ai  aj  ak  funct    A_th0       A_cth       B_th0       B_cth\n");
      }
    write_top_angles_pert(atop);
      }


    // DIHEDRALS
    fprintf(grotop, "\n[ dihedrals ]\n; ai  aj   ak  al  funct    phi0          cp      mult\n");

    write_top_dihedrals_notpert(atop);

    // then write perturbed dihedrals
    if (atop->ifpert > 0) {
      if (invert == 0) {
        fprintf(grotop, "; ai  aj   ak  al  funct  B_phi0        B_cp    B_mult      A_phi0        A_cp    A_mult\n");
      }
      else {
        fprintf(grotop, "; ai  aj   ak  al  funct  A_phi0        A_cp    A_mult      B_phi0        B_cp    B_mult\n");
      }
    write_top_dihedrals_pert(atop);
      }

    fprintf(grotop, "\n");    

}


/* write GROMACS topology file */
void write_grotop(s_atop* atop) /* --------------------------------------------- */
{
  unsigned int counter;
  /* atom types & resiue labels */
  char* type1;
  char* type2;
  char* type3;
  char* type4;
  /* atom parameters */
  unsigned int atomindex;
  char* atomname;
  char* atomtype;
  double charge;
  double mass;
  /* perturbed atom parameters */
  char* atomtype_p;
  double charge_p;
  double mass_p;
  /* total charge */
  double sum_charge;
  double sum_charge_p;
  /* residue parameters */
  unsigned int from, to;
  unsigned int nres, resindex;
  char* reslabel;

/* debug memory allocation */
#ifdef DEBUG_MEM
  mtrace();
#endif
#ifdef DEBUG_GTOP
  printf("\nopening %s...\n", grotop_filename);
#endif

/* open file */
grotop = open_file(grotop_filename, "w");

  counter = 1;

  type1 = (char*) xmalloc(5);
  type2 = (char*) xmalloc(5);
  type3 = (char*) xmalloc(5);
  type4 = (char*) xmalloc(5);
  reslabel   = (char*) xmalloc(5);
  atomname   = (char*) xmalloc(5);
  atomtype   = (char*) xmalloc(5);
  atomtype_p = (char*) xmalloc(5);

// write header
  fprintf(grotop, ";\n; Topology from AMBER\n; ambgro\n;\n; The force field files to be included\n");
  fprintf(grotop, "#include \"%s\"\n\n", groitp_filename);

/* include solvent topology */
  if (sol == 0)
    {
      fprintf(grotop, "; Include solvent topology\n");
      fprintf(grotop, "#include \"%s.itp\"\n\n", sol_filename);
    }

/*
// get label of first residue  
  reslabel = get_slice(atop->labres, reslabel, 4,4,0);
  fprintf(grotop, " [ moleculetype ]\n; name  nrexcl\n%4s   3\n\n", reslabel);
*/

  reslabel = get_slice(atop->labres, reslabel, 4,4,0);
  fprintf(grotop, " [ moleculetype ]\n; name  nrexcl\n%8s   3\n\n", "molecule");


  nres = atop->nres;

  fprintf(grotop, " [ atoms ]\n");

  if (atop->ifpert > 0)
    {
      if (invert == 0) {
        fprintf(grotop, ";  nr  B_type   resnr   residu   atom   cgnr  B_charge    B_mass   A_type A_charge    A_mass\n");
      }
      else {
        fprintf(grotop, ";  nr  A_type   resnr   residu   atom   cgnr  A_charge    A_mass   B_type B_charge    B_mass\n");
      }
    }
  else
    {
      fprintf(grotop, ";  nr    type   resnr   residu   atom   cgnr    charge      mass\n");
    }

  /* go through all residues and print all atoms in each residue */    
  for(resindex=0; resindex< nres; resindex++)
    {
      /* get residue label and bounds */
      reslabel = get_slice(atop->labres,reslabel,4,4,resindex);

      from = atop->ipres[resindex]-1;
      if (resindex == atop->nres-1)
        {
          to = atop->natom-1;
        }
      else
        {
          to = (atop->ipres[resindex+1]-1)-1;
        }

#ifdef DEBUG_GTOP
      printf("i=%5d/%5d %8d->%5d (%5d) |%s|\n", resindex, atop->nres, from,to,to-from+1,reslabel);
#endif

      sum_charge = 0.0;
      sum_charge_p = 0.0;
      
      /* write atom info */
      for ( atomindex=from; atomindex <= to; atomindex++)
        {
          /* get atom name&type */
          atomname = get_slice(atop->igraph, atomname, 4, 4, atomindex);
          atomname = repapo(atomname);
          atomtype = get_slice(atop->isymbl, atomtype, 4, 4, atomindex);

          charge = atop->chrg[atomindex]; 
          sum_charge = sum_charge + charge;
	  /* if total charge > 0.0005, set last partial charge to zero */
/*	  if (atomindex == to)
	    {
	      if ((sum_charge > 0.9e-05) && (sum_charge > 0.0)) { charge = charge - sum_charge; }
	      if ((sum_charge < -0.9e-05) && (sum_charge < 0.0)) { charge = charge + sum_charge; }
	    } */
          mass = atop->amass[atomindex];
	  
          /* write non-pert data first */
	  if (invert == 1) {
	    // if invert is off, print it out
            fprintf(grotop,"%5d      %4s%6d  %8s%8s%5d%10.5f%10.5f", counter, atomtype, resindex+1, reslabel, atomname,counter, charge, mass); 
	  }    
	  	
	  if (atop->ifpert > 0)
	    {
	      /* get per. atom info */
              atomtype_p = get_slice(atop->ismper, atomtype_p, 4, 4, atomindex);
              charge_p = atop->cgper[atomindex]; 
              sum_charge_p = sum_charge_p + charge_p;
              mass_p = atop->almper[atomindex];

//	        if (strcmp(atomtype,atomtype_p) !=0 || not_ekviv(charge,charge_p,0.00001) || not_ekviv(mass,mass_p,0.0001)) {
                  // invert if off
		  if (invert == 1) {
		    // check equivalency
	            if (strcmp(atomtype,atomtype_p) !=0 || not_ekviv(charge,charge_p,0.00001) || not_ekviv(mass,mass_p,0.0001)) {
	              fprintf(grotop,"%8s%10.5f%10.5f", atomtype_p, charge_p, mass_p);
		    }
		  }
		  else {
		    // invert is on
		    // check equivalency
	            if (strcmp(atomtype,atomtype_p) !=0 || not_ekviv(charge,charge_p,0.00001) || not_ekviv(mass,mass_p,0.0001)) {
		      fprintf(grotop,"%5d      %4s%6d  %8s%8s%5d%10.5f%10.5f", counter, atomtype_p, resindex+1, reslabel, atomname,counter, charge_p, mass_p); 
	              fprintf(grotop,"%8s%10.5f%10.5f", atomtype, charge, mass);
		    }
		    else {
		      fprintf(grotop,"%5d      %4s%6d  %8s%8s%5d%10.5f%10.5f", counter, atomtype_p, resindex+1, reslabel, atomname,counter, charge_p, mass_p); 
		    }
		  }    

	          fprintf(grotop,"\n");
	    }
	  else
	    {
	      fprintf(grotop, "\n");
	    }
          counter++;
        }
    }

  // write constraints and position_restraints
  if (constraints == 0)  { write_constraints(atop); }
  if (position_constraints == 0) { 
    fprintf(grotop,"\n#include \"%s\"\n", constraints_filename);
  }

/* write bonded information to .top */
  write_top_bon(atop);


/* system */
  fprintf(grotop, " [ system ]\n; title from amber topology\n");
  if (atop->notitle == 0)
    {
      fprintf(grotop, "%s\n\n", rst_title);
    }
  else
    {
      fprintf(grotop, "no title\n\n");
    }

/* molecules */
  fprintf(grotop, " [ molecules ]\n; molecule name    nr.\n");

/*
  for(resindex=0; resindex< nres; resindex++)
    {
      // get residue label
      reslabel = get_slice(atop->labres,reslabel, 4,4,resindex);
      
      fprintf(grotop, "%4s%16d\n", reslabel, 1);
    }
*/

      fprintf(grotop, "%8s%12d\n", "molecule", 1);


/* finalize */
close_file(grotop);

#ifdef USE_FREE
free(type1);
free(type2);
free(type3);
free(type4);
free(atomname);
free(atomtype);
free(atomtype_p);
free(reslabel);
#endif

#ifdef DEBUG_GTOP
  printf("%s closed.\n", grotop_filename);
#endif
#ifdef DEBUG_MEM
  muntrace();
#endif
}

// F.R. BON
void write_top_bon_pert(s_atop* atop) {
    unsigned int i;
    /* atom number */
    int atom1, atom2;
    unsigned int index;
    /* force constants, length, angles */
    double bfc, bl;
    double bfc_p, bl_p;
    /* atom type */
    char* type1;
    char* type2;
    //
    unsigned int is_equiv;

    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       

    is_equiv = 0;

        for (i=0; i < atop->nbper; i++)
	  {
	    /* lambda=1, non-perturbed */
	    atom1 = atop->ibper[i];
	    atom2 = atop->jbper[i];
	    atom1--;
	    atom2--;
	    index = atop->icbper[i];
            /* force constant & bond length kcal/mol,A */
	    bfc = atop->rk[index-1];
	    bl = atop->req[index-1];
            /* convert to kJ/mol, nm */
	    bfc = bfc*conv*100.000*2.000;
            bl = bl*0.1000;
            /* get atom name */
	    type1 = get_slice(atop->igraph, type1, 4,4, atom1);
            type1 = repapo(type1);

	    type2 = get_slice(atop->igraph, type2, 4,4, atom2);
            type2 = repapo(type2);
	    
	    /* lambda=0, perturbed */
	    index = atop->icbper[i+(atop->nbper)];
            /* force constant & bond length kcal/mol,A */
	    bfc_p = atop->rk[index-1];
	    bl_p = atop->req[index-1];
            /* convert to kJ/mol, nm */
	    bfc_p = bfc_p*conv*100.000*2.000;
            bl_p = bl_p*0.1000;
            
	    is_equiv = 1;
	    
            if (not_ekviv(bl,bl_p,0.00001) || not_ekviv(bfc,bfc_p,0.1)) {
	      is_equiv = 0;
	    }


            if (invert == 1) {
              fprintf(grotop, "  %6d%6d%s%12.5f%12.0f%s", ++atom1, ++atom2, "   1", bl, bfc,".");
              fprintf(grotop, "  %12.5f%12.0f%s", bl_p, bfc_p,".");
	      if (is_equiv == 1) {
                fprintf(grotop, "       ;%4s-%4s\n", type1, type2);
	      }
	      else {
                fprintf(grotop, "       ;%4s-%4s NEQ\n", type1, type2);
	      }
	    }

	    else {
              fprintf(grotop, "  %6d%6d%s%12.5f%12.0f%s", ++atom1, ++atom2, "   1", bl_p, bfc_p,".");
              fprintf(grotop, "  %12.5f%12.0f%s", bl, bfc,".");
              if (is_equiv == 1) {
                fprintf(grotop, "       ;%4s-%4s\n", type1, type2);
	      }
	      else {
                fprintf(grotop, "       ;%4s-%4s NEQ\n", type1, type2);
	      }
	    }
	    
	  }

free(type1);
free(type2);
}

// ---------------------------------------------------------------------------

void write_top_bon_notpert(s_atop* atop) {
    unsigned int i;
    /* atom number */
    int atom1, atom2;
    unsigned int index;
    /* force constants, length, angles */
    double bfc, bl;
    /* atom type */
    char* type1;
    char* type2;

    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       

    /* bonds without hydrogen */
    for (i=0; i < atop->mbona; i++)
      {
        atom1 = atop->ib[i];
	atom2 = atop->jb[i];
	atom1--;
	atom2--;
	index = atop->icb[i];
        /* force constant & bond length kcal/mol,A */
	bfc = atop->rk[index-1];
	bl = atop->req[index-1];
        /* convert to kJ/mol, nm */
	bfc = bfc*conv*200.000;
        bl = bl*0.1000;
        /* get atom name */
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
        type1 = repapo(type1);
	
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
        type2 = repapo(type2);

        fprintf(grotop, "  %6d%6d%s%12.5f%12.0f%s", ++atom1, ++atom2, "   1", bl, bfc,".");
        fprintf(grotop, "%s%4s-%4s\n", "   ;", type1, type2);
      }

    /* bonds with hydrogen */
    for (i=0; i < atop->nbonh; i++)
      {
        atom1 = atop->ibh[i];
	atom2 = atop->jbh[i];
	atom1--;
	atom2--;
	index = (atop->icbh[i]);
        /* force constant & bond length kcal/mol,A */
	bfc = atop->rk[index-1];
	bl = atop->req[index-1];
        /* convert to kJ/mol, nm */
	bfc = bfc*conv*200.000;
        bl = bl*0.1000;
        /* get atom type */
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	type1 = repapo(type1);
	
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	type2 = repapo(type2);

        fprintf(grotop, "  %6d%6d%s%12.5f%12.0f%s", ++atom1, ++atom2, "   1", bl, bfc,".");
        fprintf(grotop, "%s%4s-%4s\n", "   ;", type1, type2);
      }

free(type1);
free(type2);
}


// print pairs in perturbed dihedrals
void write_top_pairs_pert(s_atop* atop) {

int atom1, atom2, atom3, atom4;
unsigned int i;

        for (i=0; i < atop->ndper; i++)
	  {
            atom1 = atop->ipper[i];
	    atom2 = atop->jpper[i];
	    atom3 = atop->kpper[i];
	    atom4 = atop->lpper[i];
        
	    /* exclude impropers and 'end group' */
	    if (atom3 > 0 && atom4 > 0)
	      {
	        fprintf(grotop, "%8d%8d  1\n", atom1, atom4);
	      }      
	  }
}

// print pairs in perturbed dihedrals
void write_top_pairs_notpert(s_atop* atop) {

int atom1, atom2, atom3, atom4;
unsigned int i;

    for (i=0; i < atop->mphia; i++)
      {
        atom1 = atop->ip[i];
	atom2 = atop->jp[i];
	atom3 = atop->kp[i];
	atom4 = atop->lp[i];
        
	// exclude impropers and 'end group'
	if (atom3 > 0 && atom4 > 0)
	  {
	    fprintf(grotop, "%8d%8d  1\n", atom1, atom4);
	  }      
      }

    for (i=0; i < atop->nphih; i++)
      {
        atom1 = atop->iph[i];
	atom2 = atop->jph[i];
	atom3 = atop->kph[i];
	atom4 = atop->lph[i];
        
	/* exclude impropers and 'end group' */
	if (atom3 > 0 && atom4 > 0)
	  {
	    fprintf(grotop, "%8d%8d  1\n", atom1, atom4);
	  }      
      }
}



void write_top_angles_pert(s_atop* atop)
  {
    unsigned int i;
    /* atom number */
    int atom1, atom2, atom3;
    unsigned int index;
    /* force constants, angles */
    double afc, av;
    /* dtto for perturbed params */
    double afc_p, av_p;
    /* atom type */
    char* type1;
    char* type2;
    char* type3;
    //
    unsigned int is_equiv;
    
    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       
    type3 = (char*) xmalloc(5);       

        for (i=0; i < atop->ngper; i++)
	  {
	    /* first print out non-perturbed paramesters */
            atom1 = atop->ipter[i];
	    atom2 = atop->jpter[i];
	    atom3 = atop->kpter[i];
            atom1--;
	    atom2--;
	    atom3--;
	    index = atop->icpter[i];
            /* force constant in kcal/mol, deg */
	    afc = atop->tk[index-1];
	    av  = atop->teq[index-1];
            /* convert to kJ/mol, deg */
	    afc = afc*conv*2.000;
            av = av*1.000;
            /* get atom type */
	    type1 = get_slice(atop->igraph, type1, 4,4, atom1);
            type1 = repapo(type1);
	    type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	    type2 = repapo(type2);
	    type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	    type3 = repapo(type3);

            /* get parameters for perturbed types */
	    index = atop->icpter[i+(atop->ngper)];
            /* force constant in kcal/mol, deg */
	    afc_p = atop->tk[index-1];
	    av_p  = atop->teq[index-1];
            /* convert to kJ/mol, deg */
	    afc_p = afc_p*conv*2.000;
            av_p = av_p*1.000;

	    is_equiv = 1;
	    
            if (not_ekviv(av,av_p,0.0001) || not_ekviv(afc,afc_p,0.1)) {
	      is_equiv = 0;
	    }

	    
	    if (invert == 1) {
              fprintf(grotop, "%6d%6d%6d%s%12.3f%12.4f", ++atom1, ++atom2, ++atom3, "   1", av, afc);
              fprintf(grotop, "%12.3f%12.4f", av_p, afc_p);
	      if (is_equiv == 0) {
	        fprintf(grotop, "%s%4s-%4s-%4s NEQ\n", "     ;", type1, type2, type3);
	      }
	      else {
	        fprintf(grotop, "%s%4s-%4s-%4s\n", "     ;", type1, type2, type3);
	      }
	    }
	    
	    else {
              fprintf(grotop, "%6d%6d%6d%s%12.3f%12.4f", ++atom1, ++atom2, ++atom3, "   1", av_p, afc_p);
              fprintf(grotop, "%12.3f%12.4f", av, afc);
	      if (is_equiv == 0) {
                fprintf(grotop, "%s%4s-%4s-%4s NEQ\n", "     ;", type1, type2, type3);
	      }
	      else {
                fprintf(grotop, "%s%4s-%4s-%4s\n", "     ;", type1, type2, type3);
	      }
	    }
	    	    
	  }

free(type1);
free(type2);
free(type3);

}


void write_top_angles_notpert(s_atop* atop)
  {
    unsigned int i;
    /* atom number */
    int atom1, atom2, atom3;
    unsigned int index;
    /* force constants, angles */
    double afc, av;
    /* atom type */
    char* type1;
    char* type2;
    char* type3;

    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       
    type3 = (char*) xmalloc(5);       

    // angles without hydrogen
    for (i=0; i < atop->mtheta; i++)
      {
        atom1 = atop->it[i];
	atom2 = atop->jt[i];
	atom3 = atop->kt[i];
        atom1--;
	atom2--;
	atom3--;
	index = atop->ict[i];
        // force constant in kcal/mol, deg
	afc = atop->tk[index-1];
	av  = atop->teq[index-1];
        // convert to kJ/mol, deg
	afc = afc*conv*2.000;
        av = av*1.000;
        // get atom type
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	type1 = repapo(type1);
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	type2 = repapo(type2);
	type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	type3 = repapo(type3);

        fprintf(grotop, "%6d%6d%6d%s%12.3f%12.4f", ++atom1, ++atom2, ++atom3, "   1", av, afc);
        fprintf(grotop, "%s%4s-%4s-%4s\n", "     ;", type1, type2, type3);
      }
    // angles with hydrogen
    for (i=0; i < atop->ntheth; i++)
      {
        atom1 = atop->ith[i];
	atom2 = atop->jth[i];
	atom3 = atop->kth[i];
        atom1--;
	atom2--;
	atom3--;
	index = atop->ickh[i];
        // force constant in kcal/mol, deg
	afc = atop->tk[index-1];
	av  = atop->teq[index-1];
        // convert to kJ/mol, deg
	afc = afc*conv*2.000;
        av = av*1.000;
        // get atom type
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	type1 = repapo(type1);	
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	type2 = repapo(type2);	
	type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	type3 = repapo(type3);	

        fprintf(grotop, "%6d%6d%6d%s%12.3f%12.4f", ++atom1, ++atom2, ++atom3, "   1", av, afc);
        fprintf(grotop, "%s%4s-%4s-%4s\n", "     ;", type1, type2, type3);
      }

free(type1);
free(type2);
free(type3);

}


void write_top_dihedrals_pert(s_atop* atop)
  {
    unsigned int i;
    unsigned int mtd_flag;
    /* atom number */
    int atom1, atom2, atom3, atom4;
    int atom1_tmp, atom2_tmp, atom3_tmp, atom4_tmp;
    unsigned int index;
    /* force constants, length, angles */
    double phi, dfc;
    int mult;
    /* dtto for perturbed params */
    double phi_p, dfc_p;
    int mult_p;
    unsigned int eg_flag;
    char* eg="     ";
    /* atom type */
    char* type1;
    char* type2;
    char* type3;
    char* type4;
    //
    unsigned int is_equiv;
    
    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       
    type3 = (char*) xmalloc(5);       
    type4 = (char*) xmalloc(5);       

      /* first print non-perturbed data */
      for (i = 0; i < atop->ndper; i++)
        {
          atom1 = atop->ipper[i];
	  atom2 = atop->jpper[i];
	  atom3 = atop->kpper[i];
	  atom4 = atop->lpper[i];
//           if (atom4 < 0) { continue;} /* improper dih. */
             if (atom4 < 0) { atom4 = abs(atom4); }
          eg_flag = 0;
            if (atom3 < 0) { eg_flag=1; atom3=abs(atom3);} /* end group int. */
          atom1--;
          atom2--;
	  atom3--;
	  atom4--;

	  /* reset multi-termed dihedral */
	  mtd_flag = 0;
	  do
	  {
          index = atop->icpper[i];
	  if (mtd_flag > 0) index = index + mtd_flag;
          /* force constant in kcal/mol, deg */
	  dfc  = atop->pk[index-1];
	  mult = atop->pn[index-1];
	  phi  = atop->phase[index-1];
          /* convert to kJ/mol, deg */
	  dfc = dfc*conv;
          /* get atom type */
	  type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	  type1 = repapo(type1);
	  type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	  type2 = repapo(type2);
	  type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	  type3 = repapo(type3);
	  type4 = get_slice(atop->igraph, type4, 4,4, atom4);
	  type4 = repapo(type4);

          eg = "  ";
          if (eg_flag == 1) { eg = "eg"; }

        atom1_tmp = atom1 + 1;
        atom2_tmp = atom2 + 1;
        atom3_tmp = atom3 + 1;
        atom4_tmp = atom4 + 1;

        /* now print perturbed data */
	  index = atop->icpper[i+(atop->ndper)];
          if (mtd_flag > 0) index = index + mtd_flag;
          /* force constant in kcal/mol, deg */
	  dfc_p  = atop->pk[index-1];
	  mult_p = atop->pn[index-1];
	  phi_p  = atop->phase[index-1];
          /* convert to kJ/mol, deg */
	  dfc_p = dfc_p*conv;

	  is_equiv = 1;
	    
          if (not_ekviv(phi,phi_p,0.001) || not_ekviv(dfc,dfc_p,0.001) || not_ekviv(mult,mult_p,0.1)) {
	    is_equiv = 0;
	  }


          if (invert == 1) {
	    fprintf(grotop, "%6d%6d%6d%6d%s%12.3f%12.3f%10d", atom1_tmp, atom2_tmp, atom3_tmp, atom4_tmp, "   1", phi, dfc, abs(mult));
            fprintf(grotop, "%12.3f%12.3f%10d", phi_p, dfc_p, abs(mult_p));
	    if (is_equiv == 0) {
              fprintf(grotop, "%s%s%4s-%4s-%4s-%4s NEQ\n", "    ;", eg, type1, type2, type3, type4);
	    }
	    else {
              fprintf(grotop, "%s%s%4s-%4s-%4s-%4s\n", "    ;", eg, type1, type2, type3, type4);
	    }
          }
	  else {
	    fprintf(grotop, "%6d%6d%6d%6d%s%12.3f%12.3f%10d", atom1_tmp, atom2_tmp, atom3_tmp, atom4_tmp, "   1", phi_p, dfc_p, abs(mult_p));
            fprintf(grotop, "%12.3f%12.3f%10d", phi, dfc, abs(mult));
	    if (is_equiv == 0) {
              fprintf(grotop, "%s%s%4s-%4s-%4s-%4s NEQ\n", "    ;", eg, type1, type2, type3, type4);
	    }
	    else {
              fprintf(grotop, "%s%s%4s-%4s-%4s-%4s\n", "    ;", eg, type1, type2, type3, type4);
	    }  
	  }
	  
          if (mult < 0 || mult_p < 0)
	    {
	      mtd_flag++;
	      fprintf(grotop, "; mult=%d \tindex1=%d\tread next1:%d\n", mult, atop->icpper[i], (atop->icpper[i])+1); 
	      fprintf(grotop, "; mult=%d \tindex2=%d\tread next2:%d\n", mult, index, index+1);
	    }
	  else
	    {
	      mtd_flag = 0;
	    }
	    
          } while (mtd_flag > 0);

	}
  
free(type1);
free(type2);
free(type3);
free(type4);
  
}


void write_top_dihedrals_notpert(s_atop* atop)
  {
    unsigned int i;
    unsigned int mtd_flag;
    /* atom number */
    int atom1, atom2, atom3, atom4;
    unsigned int index;
    /* force constants, length, angles */
    double phi, dfc;
    int mult;
    unsigned int eg_flag;
    char* eg="     ";
    /* atom type */
    char* type1;
    char* type2;
    char* type3;
    char* type4;

    /* allocate space for atom names from igraph */    
    type1 = (char*) xmalloc(5);       
    type2 = (char*) xmalloc(5);       
    type3 = (char*) xmalloc(5);       
    type4 = (char*) xmalloc(5);       

    /* dihedrals without hydrogen */
    for (i=0; i < atop->mphia; i++)
      {
        atom1 = atop->ip[i];
	atom2 = atop->jp[i];
	atom3 = atop->kp[i];
	atom4 = atop->lp[i];

//    	     if (atom4 < 0) { continue;} /* improper dih. */
             if (atom4 < 0) { atom4 = abs(atom4); }

        eg_flag = 0;
        if (atom3 < 0) { eg_flag=1; atom3=abs(atom3);} /* end group int. */

        atom1--;
	atom2--;
        atom3--;
	atom4--;

        /* reset multi-termed dihedral */
	mtd_flag = 0;
	do { /* do while mtd_flag > 0 */

	index = atop->icp[i]; /* get index for i-th dihedral */
	
        if (mtd_flag > 0) index = index + mtd_flag; /* if multitermed dihedral, increment index */
        
	/* force constant in kcal/mol, deg */
	dfc  = atop->pk[index-1];
	mult = atop->pn[index-1];
	phi  = atop->phase[index-1];
        /* convert to kJ/mol, deg */
	dfc = dfc*conv;
        /* get atom types */
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	type1 = repapo(type1);
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	type2 = repapo(type2);
	type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	type3 = repapo(type3);
	type4 = get_slice(atop->igraph, type4, 4,4, atom4);
	type4 = repapo(type4);

        eg = "  ";
        if (eg_flag == 1) { eg = "eg"; }
        fprintf(grotop, "%6d%6d%6d%6d%s%12.3f%12.3f%10d  ;%s", ++atom1, ++atom2, ++atom3, ++atom4, "   1", phi, dfc, abs(mult), eg);
        fprintf(grotop, "%s%4s-%4s-%4s-%4s", " ", type1, type2, type3, type4);

    	    if (mult < 0) {
		mtd_flag++;
		atom1--; atom2--; atom3--; atom4--; 
		fprintf(grotop," MTD\n");
	    }
	    else {
		mtd_flag = 0;
		fprintf(grotop,"\n");
	    }
	} while (mtd_flag > 0); /* end of do cycle */

      }

    /* dihedrals with hydrogen */
    for (i=0; i < atop->nphih; i++)
      {
        atom1 = atop->iph[i];
	atom2 = atop->jph[i];
	atom3 = atop->kph[i];
	atom4 = atop->lph[i];

//        if (atom4 < 0) { continue;} /* improper dih. */
             if (atom4 < 0) { atom4 = abs(atom4); }

        eg_flag = 0;
        if (atom3 < 0) { eg_flag=1; atom3=abs(atom3);} /* end group int. */

        atom1--;
	atom2--;
        atom3--;
	atom4--;

        /* reset multi-termed dihedral */
	mtd_flag = 0;
	do { /* do while mtd_flag > 0 */

	index = atop->icph[i];

        if (mtd_flag > 0) index = index + mtd_flag; /* if multitermed dihedral, increment index */

        /* force constant in kcal/mol, deg */
	dfc  = atop->pk[index-1];
	mult = atop->pn[index-1];
	phi  = atop->phase[index-1];
        /* convert to kJ/mol, deg */
	dfc = dfc*conv;
        /* get atom type */
	type1 = get_slice(atop->igraph, type1, 4,4, atom1);
	type1 = repapo(type1);
	type2 = get_slice(atop->igraph, type2, 4,4, atom2);
	type2 = repapo(type2);
	type3 = get_slice(atop->igraph, type3, 4,4, atom3);
	type3 = repapo(type3);
	type4 = get_slice(atop->igraph, type4, 4,4, atom4);
	type4 = repapo(type4);

        eg = "  ";
        if (eg_flag == 1) { eg = "eg"; }
        fprintf(grotop, "%6d%6d%6d%6d%s%12.3f%12.3f%10d  ;%s", ++atom1, ++atom2, ++atom3, ++atom4, "   1", phi, dfc, abs(mult), eg);
        fprintf(grotop, "%s%4s-%4s-%4s-%4s", " ", type1, type2, type3, type4);

    	    if (mult < 0) {
		mtd_flag++;
		atom1--; atom2--; atom3--; atom4--; 
		fprintf(grotop," MTD\n");
	    }
	    else {
		mtd_flag = 0;
		fprintf(grotop,"\n");
	    }
	} while (mtd_flag > 0); /* end of do cycle */

      }  /* end of 'for' cycle "dihedrals with hydrogens" */

free(type1);
free(type2);
free(type3);
free(type4);
  
}
