/* AMBCONV - program for convering AMBER topology to GROMACS/TINKER topology */
/* Filip Ryjacek 2002 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <math.h>
/* #include <mcheck.h> */

#include "ambconv.h"
#include "write_tintop.h"

// write TINKER topology file
void write_tintop(s_atop* atop) {

  unsigned int counter;
  // atom types & resiue labels
  char* type1;
  char* type2;
  char* type3;
  char* type4;
  char* tree_join;
  // atom parameters
  unsigned int atomindex;
  char* atomname;
  char* atomtype;
  double mass;
  unsigned int atom_number;
  unsigned int atom_connectivity;
  // residue parameters
  unsigned int from, to;
  unsigned int nres, resindex;
  char* reslabel;
  // other
  unsigned int atoms;
  int* tin_unique_array;

// debug memory allocation
#ifdef DEBUG_MEM
  mtrace();
#endif
#ifdef DEBUG_TTOP
  printf("\nwrite_tintop: opening %s...\n", tintop_filename);
#endif

/* open file */
tintop = open_file(tintop_filename, "w");

  counter = 0;
  atoms = atop->natom;

  type1 = (char*) xmalloc(5);
  type2 = (char*) xmalloc(5);
  type3 = (char*) xmalloc(5);
  type4 = (char*) xmalloc(5);
  reslabel   = (char*) xmalloc(5);
  atomname   = (char*) xmalloc(5);
  atomtype   = (char*) xmalloc(5);
  tree_join = (char*) xmalloc(5);
  tin_unique_array = (int*) xmalloc(atoms*sizeof(int));

  // set end mark (see tin_unique)
  tin_unique_array[0] = -1;


// write header
   write_header();

// write AMBER topology title
  fprintf(tintop, "\n# ");

  if (atop->notitle == 0)
    {
      fprintf(tintop, "%s\n\n", rst_title);
    }
  else
    {
      fprintf(tintop, "no title\n\n");
    }

   
  nres = atop->nres;

// go through all residues and print all atoms in each residue
  for(resindex=0; resindex< nres; resindex++)
    {
      // get residue label and bounds
      reslabel = get_slice(atop->labres, reslabel, 4,4,resindex);

      from = atop->ipres[resindex]-1;
      if (resindex == atop->nres-1)
        {
          to = atop->natom-1;
        }
      else
        {
          to = (atop->ipres[resindex+1]-1)-1;
        }

      fprintf(tintop,"#%4s number:%4d\n", reslabel, resindex+1);

#ifdef DEBUG_TTOP
      printf("i=%5d/%5d %8d->%5d (%5d) |%s|\n", resindex, atop->nres, from,to,to-from+1,reslabel);
#endif

      // write atom info
      for ( atomindex=from; atomindex <= to; atomindex++)
        {
          // get atom name&type
          atomname = get_slice(atop->igraph, atomname, 4, 4, atomindex);

          atomtype = get_slice(atop->isymbl, atomtype, 4, 4, atomindex);
          mass = atop->amass[atomindex];
	  
          fprintf(tintop,"atom%7d%6d%6s ", counter+1, atop->iac[counter] ,atomtype);
	  
	  // calculate atom number
            atom_number = floor(mass/2);
	    if ( atom_number < 1) {
	      atom_number = 1;
	    }

	  fprintf(tintop,"\"%4s:%4s\"    %10d%10.3f", reslabel, atomname, atom_number, mass);
	  
	  tree_join = get_slice(atop->itree, tree_join, 4, 4, counter);
	  atom_connectivity = tree_to_connectivity(tree_join, counter, atoms);
	  
	  fprintf(tintop," %2d", atom_connectivity); 
	  fprintf(tintop, "\n");
	  counter++;
        }
    }

counter = 0;

  fprintf(tintop,"\n# Van der Waals Parameters\n\n");
  // vdw information will write write_itp() function


#ifdef USE_FREE
free(type1);
free(type2);
free(type3);
free(type4);
free(atomname);
free(atomtype);
free(reslabel);
free(tree_join);
free(tin_unique_array);
#endif

/* finalize */
close_file(tintop);
#ifdef DEBUG_TTOP
  printf("write_tintop: %s closed.\n", tintop_filename);
#endif

#ifdef DEBUG_MEM
  muntrace();
#endif

}


void write_header(void) {
  fprintf(tintop,"\n");
  fprintf(tintop,"   ##############################\n");
  fprintf(tintop,"   ##                          ##\n");
  fprintf(tintop,"   ##  Force Field Definition  ##\n");
  fprintf(tintop,"   ##                          ##\n");
  fprintf(tintop,"   ##############################\n");
  fprintf(tintop,"\n\n");
  fprintf(tintop,"forcefield   AMBER94\n\n"); 
  fprintf(tintop,"vdwtype      LENNARD-JONES\n"); 
  fprintf(tintop,"radiusrule   ARITHMETIC\n"); 
  fprintf(tintop,"radiustype   R-MIN\n"); 
  fprintf(tintop,"radiussize   RADIUS\n"); 
  fprintf(tintop,"epsilonrule  GEOMETRIC\n"); 
  fprintf(tintop,"vdw-14-scale 2.0\n"); 
  fprintf(tintop,"chg-14-scale 1.2\n"); 
  fprintf(tintop,"dielectric   1.0\n"); 
  fprintf(tintop,"\n"); 
  fprintf(tintop,"\n");
  fprintf(tintop,"   #############################\n");
  fprintf(tintop,"   ##                          ##\n");
  fprintf(tintop,"   ##  Atom Type Definitions   ##\n");
  fprintf(tintop,"   ##                          ##\n");
  fprintf(tintop,"   #############################\n");
  fprintf(tintop,"\n\n");
  fprintf(tintop,"   #################################################\n");
  fprintf(tintop,"   #                                               #\n");
  fprintf(tintop,"   # TINKER Atom Class Numbers to Amber Atom Types #\n\n");

  fprintf(tintop,"   #                                               #\n");
  fprintf(tintop,"   #################################################\n");
  
}

// converts tree joining info to atom connectivity
unsigned int tree_to_connectivity(char* tree_join, unsigned int counter, unsigned int atoms) {
  unsigned int atom_connectivity;
  
  // default value
  atom_connectivity = 1;

  // main
  if (strcmp(tree_join,"M   ") == 0) {
    atom_connectivity = 2;
  }

  // branch
  if (strcmp(tree_join,"B   ") == 0) {
    atom_connectivity = 3;
  }

  // triple-fork
  if (strcmp(tree_join,"3   ") == 0) {
    atom_connectivity = 4;
  }
  
  // first or last atom
  if (counter == 0 || counter == (atoms-1)) {
    atom_connectivity = 1;
  }
  
  // end-atom
  if (strcmp(tree_join,"E   ") == 0) {
    atom_connectivity = 1;
  }
  
  return atom_connectivity;
}

// if vdw atom type is unique, return null
unsigned int tin_unique(unsigned int atomindex, int* tin_unique_array, s_atop* atop) {
  unsigned int unique;
  unsigned int counter;
  
  unique = 0;
  counter = 0;

#ifdef DEBUG_TTOP
  printf("\n\n");
#endif
  
  // searching in array...
  do {

#ifdef DEBUG_TTOP
  printf("TU: [%d] comparing %d with %d\n", counter, tin_unique_array[counter], atop->iac[atomindex]);
#endif
  
    if (tin_unique_array[counter] == atop->iac[atomindex]) {
#ifdef DEBUG_TTOP
      printf ("TU: is NOT unique!\n");
#endif

      unique = 1;
      break;
    }
    counter++;
  } while (tin_unique_array[counter-1] >= 0);
  
  // vdw not found, i.e. is unique
  if (unique == 0) {
    tin_unique_array[counter] = -1;

#ifdef DEBUG_TTOP
    printf("TU: unique vdw found, placing -1 at position %d\n", counter);
#endif

    tin_unique_array[counter-1] = atop->iac[atomindex];

#ifdef DEBUG_TTOP
    printf("TU: and %d at %d\n", atop->iac[atomindex], counter-1);
#endif

  }
  
  return unique;
}
